/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 * George Konstantoulakis <gkon@inaccessnetworks.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Say numbers and dates (maybe words one day too)
 * 
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>

#ifdef SOLARIS
#include <iso/limits_iso.h>
#endif

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 1.73 $")

#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/logger.h"
#include "asterisk/options.h"
#include "asterisk/say.h"
#include "asterisk/lock.h"
#include "asterisk/localtime.h"
#include "asterisk/utils.h"

/*
 * Useful constants used to parse numbers and dates
 */
#define	ONE_HUNDRED		100
#define	ONE_THOUSAND	1000
#define	ONE_MILLION		1000000
#define	ONE_BILLION		1000000000

#define	ONE_DAY			86400	/* seconds in one day */
#define	SIX_DAYS		(ONE_DAY * 6)	/* seconds in six day */

/*
 * Typical 'say' arguments in addition to the date or number or string
 * to say. We do not include 'options' because they may be different
 * in recursive calls, and so they are better left as an external
 * parameter.
 */
typedef struct {
	struct ast_channel *chan;
	const char *ints;
	const char *language;
	int audiofd;
	int ctrlfd;
} say_args_t;

/*
 * The code uses 4 different methods to play a file.
 * They can be probably reduced to 2 or even just one,
 * but for the time being let's stick with the status quo.
 */
static int s_streamit(say_args_t *a, const char *fn)
{
	int res = ast_streamfile(a->chan, fn, a->language);

	if (!res) 
		res = ast_waitstream_full(a->chan, a->ints, a->audiofd, a->ctrlfd);
	ast_stopstream(a->chan);
	return res;
}

static int streamwait(struct ast_channel *chan, const char *fn,
	const char *lang, const char *ints)
{
	int res = ast_streamfile(chan, fn, lang);
	if (!res)
		res = ast_waitstream(chan, ints);
	return res;
}

static int s_streamwait2(say_args_t *a, const char *fn)
{
	int res = ast_streamfile(a->chan, fn, a->language);

	if (!res) {
		res = (a->audiofd  > -1 && a->ctrlfd > -1) ?
			ast_waitstream_full(a->chan, a->ints, a->audiofd, a->ctrlfd) :
			ast_waitstream(a->chan, a->ints);
	}
	return res;
}

static int s_streamwait3(say_args_t *a, const char *fn)
{
	int res = s_streamwait2(a, fn);
	ast_stopstream(a->chan);
	return res;
}

static int wait_file(struct ast_channel *chan, const char *ints, const char *file, const char *lang) 
{
	int res = streamwait(chan, file, lang, ints);
	if (res)
		ast_log(LOG_WARNING, "Unable to play message %s\n", file);
	return res;
}

static int s_waitfile(say_args_t *a, const char *fn)
{
	int res = ast_streamfile(a->chan, fn, a->language);
	if (res)
		ast_log(LOG_WARNING, "Unable to play message %s\n", fn);
	else
		res = ast_waitstream(a->chan, a->ints);
	return res;
}

/*
 * The basic say APIs has these 8 calls:
	ast_say_{character_str|phonetic_strdigit_str|digirs}[_full]

	character_str	spell out individual characters
	phonetic_str	spell out phonetic chars (maybe useful for pronounce ?)
	digit_str		spell out digit/dialpad symbols
	digits			print the number and spell out the digits
	
	full:	monitor audiofd and ctrlfd, otherwise set both to-1 

    Internally, we pack arguments into a say_args_t structure, and then
    call the actual implementation.

    They use an internal API to simplify the rest
 */

static int say_character_str_full(say_args_t *a, const char *str)
{
	char c, fnbuf[256];
	int num = 0;
	int res = 0;

	for (num = 0; (c = str[num]) != '\0'; num++) {
		const char *fn = NULL;
		switch (c) {
		case '*':
			fn = "digits/star";
			break;
		case '#':
			fn = "digits/pound";
			break;
		case '!':
			fn = "letters/exclaimation-point";
			break;
		case '@':
			fn = "letters/at";
			break;
		case '$':
			fn = "letters/dollar";
			break;
		case '-':
			fn = "letters/dash";
			break;
		case '.':
			fn = "letters/dot";
			break;
		case '=':
			fn = "letters/equals";
			break;
		case '+':
			fn = "letters/plus";
			break;
		case '/':
			fn = "letters/slash";
			break;
		case ' ':
			fn = "letters/space";
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			strcpy(fnbuf, "digits/X");
			fnbuf[7] = c;
			fn = fnbuf;
			break;
		default:
			strcpy(fnbuf, "letters/X");
			fnbuf[8] = tolower(c);	/* file names are lower-case */
			fn = fnbuf;
		}
		res = s_streamit(a, fn);
	}

	return res;
}

static int say_phonetic_str_full(say_args_t *a, const char *str)
{
	char c, fnbuf[256];
	int num = 0;
	int res = 0;

	for (num = 0; (c = str[num]) != '\0'; num++) {
		const char *fn = NULL;
		switch (c) {
		case '*':
			fn = "digits/star";
			break;
		case '#':
			fn = "digits/pound";
			break;
		case '!':
			fn = "letters/exclaimation-point";
			break;
		case '@':
			fn = "letters/at";
			break;
		case '$':
			fn = "letters/dollar";
			break;
		case '-':
			fn = "letters/dash";
			break;
		case '.':
			fn = "letters/dot";
			break;
		case '=':
			fn = "letters/equals";
			break;
		case '+':
			fn = "letters/plus";
			break;
		case '/':
			fn = "letters/slash";
			break;
		case ' ':
			fn = "letters/space";
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
			strcpy(fnbuf, "digits/X");
			fnbuf[7] = c;
			fn = fnbuf;
			break;
		default:	/* '9' falls here... */
			strcpy(fnbuf, "phonetic/X_p");
			fnbuf[9] = tolower(c);
			fn = fnbuf;
		}
		res = s_streamit(a, fn);
	}
	return res;
}

static int say_digit_str_full(say_args_t *a, const char *str)
{
	char c, fnbuf[256];
	int num = 0;
	int res = 0;

	for (num = 0; (c = str[num]) != '\0' && res == 0; num++) {
		const char *fn = NULL;
		switch (c) {
		case '*':
			fn = "digits/star";
			break;
		case '#':
			fn = "digits/pound";
			break;
		case '-':
			fn = "digits/minus";
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			strcpy(fnbuf, "digits/X");
			fnbuf[7] = c;
			fn = fnbuf;
			break;
		}
		if (fn)
			res = s_streamit(a, fn);
	}
	return res;
}


static int say_digits_full(say_args_t *a, int num)
{
	char fn2[256];

	snprintf(fn2, sizeof(fn2), "%d", num);
	return say_digit_str_full(a, fn2);
}

/*
 * Ok here are the external functions
 */
int ast_say_character_str_full(struct ast_channel *chan,
	const char *str, const char *ints, const char *lang,
	int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return say_character_str_full(&a, str);
}

int ast_say_character_str(struct ast_channel *chan, const char *str,
	const char *ints, const char *lang)
{
	say_args_t a = { chan, ints, lang, -1, -1 };
	return say_character_str_full(&a, str);
}

int ast_say_phonetic_str_full(struct ast_channel *chan, const char *str,
	const char *ints, const char *lang, int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return say_phonetic_str_full(&a, str);
}

int ast_say_phonetic_str(struct ast_channel *chan, const char *str,
	const char *ints, const char *lang)
{
	say_args_t a = { chan, ints, lang, -1, -1 };
	return say_phonetic_str_full(&a, str);
}

int ast_say_digit_str_full(struct ast_channel *chan, const char *str,
	const char *ints, const char *lang, int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return say_digit_str_full(&a, str);
}

int ast_say_digit_str(struct ast_channel *chan, const char *str,
	const char *ints, const char *lang)
{
	say_args_t a = { chan, ints, lang, -1, -1 };
	return say_digit_str_full(&a, str);
}
int ast_say_digits_full(struct ast_channel *chan, int num,
	const char *ints, const char *lang, int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return say_digits_full(&a, num);
}

int ast_say_digits(struct ast_channel *chan, int num,
	const char *ints, const char *lang)
{
	say_args_t a = { chan, ints, lang, -1, -1 };
	return say_digits_full(&a, num);
}

/* Syntaxes supported, not really language codes.
      da    - Danish
      de    - German
      en    - English (US)
      en_GB - English (British)
      es    - Spanish, Mexican
      fr    - French
      he    - Hebrew
      it    - Italian
      nl    - Dutch
      no    - Norwegian
      pl    - Polish       
      pt    - Portuguese
      se    - Swedish
      tw    - Taiwanese
      ru    - Russian

 Gender:
 For Some languages the numbers differ for gender and plural
 Use the option argument 'f' for female, 'm' for male and 'n' for neuter in languages like Portuguese, French, Spanish and German.
 use the option argument 'c' is for commune and 'n' for neuter gender in nordic languages like Danish, Swedish and Norwegian.
 use the option argument 'p' for plural enumerations like in German
 
 Date/Time functions currently have less languages supported than saynumber().

 Note that in future, we need to move to a model where we can differentiate further - e.g. between en_US & en_UK

 See contrib/i18n.testsuite.conf for some examples of the different syntaxes

 Portuguese sound files needed for Time/Date functions:
 pt-ah
 pt-ao
 pt-de
 pt-e
 pt-ora
 pt-meianoite
 pt-meiodia
 pt-sss

 Spanish sound files needed for Time/Date functions:
 es-de
 es-el

 Italian sound files needed for Time/Date functions:
 ore-una
 ore-mezzanotte

*/

/*
 * Each language is expected to implement a few parsing
 * functions to say cardinal and ordinal numbers, and
 * dates and times in various formats.
 *
 * The interface for all of these is
 *	- common arguments		say_args_t *a
 *	- the number or date	'int num' or 'time_t t'
 *	- options, e.g. gender	const char *options
 *
 *	(we have not completed the transition yet).
 *
 * In practice one should try to implement functions() in
 * the following order of importance:
 *		say_number_full()
 *			cardinal numbers, used in many places
 *
 *		say_enumeration_full()
 *			cardinal numbers, used often to pronounce dates
 *
 *		say_date_with_format()
 *			the generic date and time parser.
 *
 * All others (say_date(), say_time(), say_datetime(), say_datetime_from_now()
 * can and should be implemented as calls to say_date_with_format(),
 * using a specific format.
 */

/*
 * types for the various functions
 */
typedef int say_number_full_f(say_args_t *a, int num, const char *opts);

typedef int say_enumeration_full_f(say_args_t *a, int num, const char *opts);

/*
 * in say_date_with_format_f, last arguments is seconds after
 * 00:00:00 of today, which is useful to avoid calling localtime*()
 * within the individual functions.
 */
typedef int say_date_with_format_f(say_args_t *a, struct tm *tm1,
	const char *format, time_t delta);

typedef int say_datetime_from_now_f(say_args_t *a, time_t t);

/*
 * For each language, we are supposed to fill a descriptor
 * with the language string and pointers to the various
 * functions or formats.
 */
struct _lang_sw {
	char *language;

	int initialized;			/* 0 at beginning */
	say_number_full_f		*say_number_full;
	say_enumeration_full_f	*say_enumeration_full;
	say_date_with_format_f	*say_date_with_format;
	say_datetime_from_now_f	*say_datetime_from_now;
	/* The following are implemented as formats for say_date_with_format_f */
	const char				*time_fmt;
	const char 				*datetime_fmt;
	const char 				*date_fmt;
};

/*
 * Pointers to the various descriptors are collected into
 * an array which is then used to demux depending on the
 * requested language. The first descriptor must be the
 * default one and it is assumed to have all function non-NULL.
 *
 * XXX The demux routine also takes care to replace NULL
 * fields with default values.
 */

static struct _lang_sw *map[];

static struct _lang_sw *lang_lookup(const char *language)
{
	/* XXX make sure the default is the first one */
	int i;

	for (i = 0; map[i] != NULL; i++)
		if (!strcasecmp(map[i]->language, language))
			break;
	if (map[i ]== NULL)
		i = 0;	/* default entry */
	if (i != 0 && map[i]->initialized == 0) {
#define SET(f)	{ if (map[i]->f == NULL) map[i]->f = map[0]->f; }
		/* XXX should lock ? */
		SET(say_number_full);
		SET(say_enumeration_full);
		SET(say_date_with_format);
		SET(say_datetime_from_now);

		SET(date_fmt);
		SET(time_fmt);
		SET(datetime_fmt);
#undef SET
		map[i]->initialized = 1;
	}
	return map[i];
}

/*
 * demux routines that implement the main API and call the
 * language-specific functions.
 */

/*! \brief  ast_say_number_full: call language-specific functions */
/* Called from AGI */
int ast_say_number_full(struct ast_channel *chan, int num,
	const char *ints, const char *lang, const char *options,
	int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return lang_lookup(lang)->say_number_full(&a, num, options);
}

/*! \brief  ast_say_number: call language-specific functions
 without file descriptors
*/
int ast_say_number(struct ast_channel *chan, int num,
	const char *ints, const char *lang, const char *opts)
{
	return ast_say_number_full(chan, num, ints, lang, opts, -1, -1);
}

/*! \brief  ast_say_enumeration_full: call language-specific functions */
/* Called from AGI */
int ast_say_enumeration_full(struct ast_channel *chan, int num,
	const char *ints, const char *lang, const char *opts,
	int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return lang_lookup(lang)->say_enumeration_full(&a, num, opts);
}

/*! \brief  ast_say_enumeration: call language-specific functions without file descriptors */
int ast_say_enumeration(struct ast_channel *chan, int num,
	const char *ints, const char *lang, const char *options)
{
	return ast_say_enumeration_full(chan, num, ints, lang, options, -1, -1);
}

int ast_say_date_with_format(struct ast_channel *chan,
	time_t t, const char *ints, const char *lang,
	const char *format, const char *timezone)
{
	struct tm tm, tmnow;
	time_t sec_now, beg_today;
	say_args_t a = { chan, ints, lang, -1, -1 };

	ast_localtime(t, &tm, timezone);	/* compute tm for t */
	sec_now = time(NULL);				/* compute current second */
	ast_localtime(sec_now, &tmnow, timezone);	/* convert to tm */
	/*
	 * compute the initial second of today.
	 * This might be slightly off, if we transcend a leap second,
	 * but never more off than 1 second.
	 * In any case, it saves not having to do ast_mktime()
	 * XXX this needs to be fixed, probably.
	 */
	beg_today = sec_now -
		(tmnow.tm_hour * 3600 + tmnow.tm_min * 60 + tmnow.tm_sec);

	return lang_lookup(lang)->say_date_with_format(&a, &tm,
			format, t - beg_today);
}

int ast_say_date(struct ast_channel *chan, time_t t,
	const char *ints, const char *lang)
{
	const char *fmt = lang_lookup(lang)->date_fmt;
	return ast_say_date_with_format(chan, t, ints, lang, fmt, NULL);
}

int ast_say_time(struct ast_channel *chan, time_t t,
	const char *ints, const char *lang)
{
	const char *fmt = lang_lookup(lang)->time_fmt;
	return ast_say_date_with_format(chan, t, ints, lang, fmt, NULL);
}

int ast_say_datetime(struct ast_channel *chan, time_t t,
	const char *ints, const char *lang)
{
	const char *fmt = lang_lookup(lang)->datetime_fmt;
	return ast_say_date_with_format(chan, t, ints, lang, fmt, NULL);
}

int ast_say_datetime_from_now(struct ast_channel *chan, time_t t,
	const char *ints, const char *lang)
{
	say_args_t a = { chan, ints, lang, -1, -1 };
	return lang_lookup(lang)->say_datetime_from_now(&a, t);
}


/*
 * Here we have a few utility functions used in the various parsing
 * functions, such as reliably negating a number, playing a
 * file specified in a format string, converting hours from 24
 * to 12 hours, determining am or pm, and so on.
 */

/* negates the number if possible, otherwise returns 0 */
static int negate(int num)
{
	return (num > INT_MIN) ? -num : 0;
}

/*
 * extract a literal string from the format, play the file
 * Return the next offset in *offset, and dtmf error code
 */
static int play_literal(struct ast_channel *chan, const char *ints,
	const char *lang, const char *src, int *offset)
{
	int ofs = *offset, sndoffset = 0;
	char sndfile[256];

	/* Literal name of a sound file */
	for (;;) {
		char c = src[ ++ofs ];
		if (c == '\0') {	/* string not properly terminated */
			ofs--;
			break;
		}
		if (c == '\'')
			break;
		if (sndoffset < sizeof(sndfile) - 1) {
			sndfile[sndoffset++] = c;
		} else {
			/* XXX should warn - name has been truncated */
		}
	}
	sndfile[sndoffset] = '\0';
	*offset = ofs;	/* update return value */
	return wait_file(chan, ints, sndfile, lang);
}

/*
 * changes 0 to 12, 13..23 to 1..11, which is a very common mapping
 * between 24 and 12 hour clock format.
 */
static int hr_to_12(int h)
{
	if (h == 0)
		return 12;
	else
		return (h <= 12) ? h : h - 12;
}

/*
 * given a 24-hr hour, returns a pointer to the corresponding  am/pm string
 */
static const char *hr_to_am_pm(int hour)
{
	return hour >= 12 ? "digits/p-m": "digits/a-m";
}

/*
 * Finally, we have here the main parsing routines for the
 * default language (english).
 */

/*! \brief  say_number_full_en: English syntax */
static int say_number_full_en(say_args_t *a, int num, const char *options)
{
	int res = 0;

	/* all language have a special rule to prononuce the number 0 */
	if (!num) 
		return say_digits_full(a, 0);

	/*
	 * Then enter a loop where you split the number, fill a string
	 * with the filename to play, and possibly record in additional
	 * variables any leftover components (e.g. "hundred") that you
	 * need to play before exiting the loop.
	 * Typically terminate when the number reaches 0 or any of the
	 * play commands returns a non-zero result (error or interrupted
	 * by a keypress if enabled).
	 */
	while (!res && num) {
		const char *fn = NULL;
		char buf[256] = "";

		if (num < 0) {
			fn = "digits/minus";
			num = negate(num);
		} else	if (num < 20) {
			snprintf(buf, sizeof(buf), "digits/%d", num);
			fn = buf;
			num = 0;
		} else	if (num < ONE_HUNDRED) {
			snprintf(buf, sizeof(buf), "digits/%d", num - (num%10) );
			fn = buf;
			num = num % 10;
		} else if (num < ONE_THOUSAND) {
			res = say_number_full_en(a, num / ONE_HUNDRED, options);
			if (res)
				break;
			fn = "digits/hundred";
			num %= ONE_HUNDRED;
		} else if (num < ONE_MILLION) { /* 1,000,000 */
			res = say_number_full_en(a, num / ONE_THOUSAND, options);
			if (res)
				break;
			fn = "digits/thousand";
			num = num % ONE_THOUSAND;
		} else if (num < ONE_BILLION) {	/* 1,000,000,000 */
			res = say_number_full_en(a, num / ONE_MILLION, options);
			if (res)
				break;
			fn = "digits/million";
			num = num % ONE_MILLION;
		} else {
			ast_log(LOG_DEBUG, "Number '%d' is too big for me\n", num);
			break;
		}
		res = s_streamwait3(a, fn);	/* fn is always set to something good */
	}
	return res;
}


/*! \brief  say_enumeration_full_en: English syntax */
static int say_enumeration_full_en(say_args_t *a, int num, const char *options)
{
	/*
	 * t is set when we have some high order component
	 * that force us to pronounce also leading ones.
	 */
	int res = 0, t = 0;

    /*
     * We do not pronounce something like " 0'th "
     */
	while(!res && num) {
		const char *fn;
		char buf[256] = "";

		if (num < 0) {
			/* kind of senseless for enumerations... */
			fn = "digits/minus";
			num = negate(num);
		} else if (num < 20) {
			snprintf(buf, sizeof(buf), "digits/h-%d", num);
			fn = buf;
			num = 0;
		} else if (num < ONE_HUNDRED) { 
			int tens = num / 10;
			num = num % 10;
			/* if remainder is 0 we need the ordinal for the tens */
			snprintf(buf, sizeof(buf), "digits/%s%d", num ? "" : "h-", tens*10);
			fn = buf;
		} else if (num < ONE_THOUSAND) {
			int hundreds = num / ONE_HUNDRED;
			num = num % ONE_HUNDRED;
			/*
			 * if we have a pending high component, or > 199,
			 * do pronounce the number of hundteds.
			 */
			if (hundreds > 1 || t == 1)
				res = say_number_full_en(a, hundreds, options);
			if (res)
				break;
			fn = num ? "digits/hundred" : "digits/h-hundred";
		} else if (num < ONE_MILLION) {
			int thousands = num / ONE_THOUSAND;
			num = num % ONE_THOUSAND;
			if (thousands > 1 || t == 1)
				res = say_number_full_en(a, thousands, options);
			if (res)
				break;
			fn = num ? "digits/thousand" : "digits/h-thousand";
			t = 1;
		} else if (num < ONE_BILLION) {
			int millions = num / ONE_MILLION;
			num = num % ONE_MILLION;
			res = say_number_full_en(a, millions, options);
			if (res)
				break;
			fn = num ? "digits/million" : "digits/h-million";
			t = 1;
		} else if (num < INT_MAX) {
			int billions = num / ONE_BILLION;
			num = num % ONE_BILLION;
			t = 1;
			res = say_number_full_en(a, billions, options);
			if (res)
				break;
			fn = num ? "digits/billion" : "digits/h-billion";
		} else if (num == INT_MAX) {
			fn = "digits/h-last";
			num = 0;
		} else {
			ast_log(LOG_DEBUG, "Number '%d' is too big for me\n", num);
			break;
		}
		res = s_streamwait3(a, fn);
	}
	return res;
}


static int say_date_with_format_en(say_args_t *a, struct tm *tm1,
	const char *format, time_t delta)
{
	struct tm tm = *tm1;
	int res=0, offset;
	char c, nextmsg[256];
	say_date_with_format_f *f = say_date_with_format_en; /* shorthand for recursive call */

	for (offset=0 ; res == 0 && (c = format[offset]) != '\0' ; offset++) {
		ast_log(LOG_WARNING, "Parsing %c (offset %d) in %s\n", format[offset], offset, format);
		switch (c) {
		/*
		 * format options are mostly those defined in strftime,
		 * with only a few exceptions namely:
		 *   - spaces and tabs are ignored;
		 *   - a string in single quotes is a filename
		 *
		 * if you add more options here, please try to be consistent
		 * with strftime(3)
		 */
		case '\'': /* Literal name of a sound file */
			res = play_literal(a->chan, a->ints, a->language, format, &offset);
			break;

		case 'A':
		case 'a':
			/* national representation of the weekday name.
			 * We do not distinguish between full (A) and abbreviated (a).
			 * Sunday - Saturday
			 */
			snprintf(nextmsg,sizeof(nextmsg), "digits/day-%d", tm.tm_wday);
			res = s_waitfile(a, nextmsg);
			break;

		case 'B':
		case 'b':
		case 'h':
			/*
			 * national representation of the month name.
			 * Again do not distinguish between full (B) and abbreviated (b or h)
			 * 
			 * January - December
			 */
			snprintf(nextmsg,sizeof(nextmsg), "digits/mon-%d", tm.tm_mon);
			res = s_waitfile(a, nextmsg);
			break;

		case 'm':
			/* Month enumerated */
			res = say_enumeration_full_en(a, tm.tm_mon + 1, NULL);	
			break;

		case 'd':
		case 'e':
			/* ordinal day of month First - Thirtyfirst */
			res = say_enumeration_full_en(a, tm.tm_mday, NULL);	
			break;

		case 'Y':
			/* Year */
			if (tm.tm_year < 1) {
				/* I'm not going to handle 1900 and prior */
				/* We'll just be silent on the year, instead of bombing out. */
			} else if (tm.tm_year > 99) {	/* two thousand and more */
				res = say_number_full_en(a, tm.tm_year + 1900, NULL);
			} else {
				res = s_waitfile(a, "digits/19");
				if (res)
					break;
				if (tm.tm_year <= 9) /* 1901 - 1909 */
					res = s_waitfile(a, "digits/oh");
				if (!res)
					res = say_number_full_en(a, tm.tm_year, NULL);
			}
			break;

		case 'I':
		case 'l':
			/* 12-Hour */
			res = say_number_full_en(a, hr_to_12(tm.tm_hour), NULL);
			break;

		case 'H': /* leading 'oh' for 0..9 */
		case 'k': /* pronounce 0 as 'oh' */
			/* 24-Hour */
			if ( tm.tm_hour == 0 || (c == 'H' && tm.tm_hour < 10))
				res = s_waitfile(a, "digits/oh");
			if (res == 0 && tm.tm_hour > 0)
				res = say_number_full_en(a, tm.tm_hour, NULL);
			break;

		case 'M':
		case 'N':
			/* Minute */
			if (tm.tm_min == 0) { /* M -> 'oclock', N -> 'hundred' */
				res = s_waitfile(a, c=='M' ? "digits/oclock" : "digits/hundred");
			} else {
				if (tm.tm_min < 10)
					res = s_waitfile(a, "digits/oh");
				if (!res)
					res = say_number_full_en(a, tm.tm_min, NULL);
			}
			break;

		case 'P':
		case 'p':
			/* AM/PM */
			res = s_waitfile(a, hr_to_am_pm(tm.tm_hour));
			break;

		case 'Q': /* Shorthand for "Today", "Yesterday", or ABdY */
		case 'q': /* Shorthand for "" (today), "Yesterday", A (weekday), or ABdY */
			if (delta >= 0) { /* Today */
				if (c == 'Q')
					res = s_waitfile(a, "digits/today");
			} else if (delta > -ONE_DAY) /* Yesterday */
				res = s_waitfile(a, "digits/yesterday");
			else if (c == 'q' && delta > - SIX_DAYS) /* Within the week */
				res = f(a, &tm, "A", delta);
			else	/* full date */
				res = f(a, &tm, "ABdY", delta);
			break;
		case 'R': /* equivalent to hour minute */
			res = f(a, &tm, "HM", delta);
			break;

		case 'S': /* Seconds */
			if (tm.tm_sec > 0 && tm.tm_sec < 10)
				res = s_waitfile(a, "digits/oh");
			if (!res)
				res = say_number_full_en(a, tm.tm_sec, NULL);
			break;

		case 'T':	/* equivalent to hour, minute, second */
			res = f(a, &tm, "HMS", delta);
			break;

		case ' ':
		case '\t':
			/* Just ignore spaces and tabs */
			break;

		default:
			/* Unknown character */
			ast_log(LOG_WARNING, "Unknown character in datetime format %s: %c at pos %d\n", format, c, offset);
		}
	}
	return res;
}

/* XXX double check this */
static int say_datetime_from_now_en(say_args_t *a, time_t t)
{
	int res=0;
	time_t nowt;
	int daydiff;
	struct tm tm;
	struct tm now;
	char fn[256];

	ast_log(LOG_WARNING, "say_datetime_from_now_en lang '%s' ints '%s'",
		a->language, a->ints);
	time(&nowt);

	localtime_r(&t,&tm);
	localtime_r(&nowt,&now);
	daydiff = now.tm_yday - tm.tm_yday;
	if (daydiff < 0 || daydiff > 6) {
		/* Day of month and month */
		if (!res) {
			snprintf(fn, sizeof(fn), "digits/mon-%d", tm.tm_mon);
			res = s_waitfile(a, fn);
		}
		if (!res)
			res = say_number_full_en(a, tm.tm_mday, NULL);

	} else if (daydiff) {
		/* Just what day of the week */
		if (!res) {
			snprintf(fn, sizeof(fn), "digits/day-%d", tm.tm_wday);
			res = s_waitfile(a, fn);
		}
	} /* Otherwise, it was today */
	if (!res)
		res = ast_say_time(a->chan, t, a->ints, a->language);
	return res;
}

/************ END ENGLISH ***************************/

/*
 * Here we include all the language-specific versions,
 * and build the global map.
 *
 * XXX of course a better way should be implemented!
 */

#include "say/say_cz.c"
#include "say/say_da.c"
#include "say/say_de.c"
#include "say/say_en_GB.c"
#include "say/say_es.c"
#include "say/say_fr.c"
#include "say/say_gr.c"
#include "say/say_he.c"
#include "say/say_it.c"
#include "say/say_nl.c"
#include "say/say_no.c"
#include "say/say_pl.c"
#include "say/say_pt.c"
#include "say/say_ru.c"
#include "say/say_se.c"
#include "say/say_tw.c"

static struct _lang_sw map_en = {
	.language = "en",
	.initialized = 0,
	.say_number_full = say_number_full_en,
	.say_enumeration_full = say_enumeration_full_en,
	.say_date_with_format = say_date_with_format_en,
	.say_datetime_from_now = say_datetime_from_now_en,
	.date_fmt = "ABY",
	.time_fmt = "IMP",
	.datetime_fmt = "ABdIMP",
};

static struct _lang_sw *map[] = {
	&map_en,
	&map_cz,
	&map_da,
	&map_de,
	&map_en_GB,
	&map_es,
	&map_fr,
	&map_gr,
	&map_he,
	&map_it,
	&map_nl,
	&map_no,
	&map_pl,
	&map_pt,
	&map_ru,
	&map_se,
	&map_tw,
	NULL };

