/*
 * say_da.c
 */

/*! \brief  say_number_full_da: Danish syntax */
/* New files:
 In addition to English, the following sounds are required: "1N", "millions", "and" and "1-and" through "9-and" 
 */
static int say_number_full_da(say_args_t *a, int num, const char *options)
{
	int res = 0;
	int playh = 0;
	int playa = 0;
	int cn = 1;		/* +1 = commune; -1 = neuter */
	char fn[256] = "";
	say_number_full_f *f = say_number_full_da;
	if (!num) 
		return say_digits_full(a, 0);

	if (options && !strncasecmp(options, "n",1))
		cn = -1;

	while(!res && (num || playh || playa )) {
		/* The grammar for Danish numbers is the same as for English except
		* for the following:
		* - 1 exists in both commune ("en", file "1N") and neuter ("et", file "1")
		* - numbers 20 through 99 are said in reverse order, i.e. 21 is
		*   "one-and twenty" and 68 is "eight-and sixty".
		* - "million" is different in singular and plural form
		* - numbers > 1000 with zero as the third digit from last have an
		*   "and" before the last two digits, i.e. 2034 is "two thousand and
		*   four-and thirty" and 1000012 is "one million and twelve".
		*/
		if (num < 0) {
			snprintf(fn, sizeof(fn), "digits/minus");
			num = negate(num);
		} else if (playh) {
			snprintf(fn, sizeof(fn), "digits/hundred");
			playh = 0;
		} else if (playa) {
			snprintf(fn, sizeof(fn), "digits/and");
			playa = 0;
		} else if (num == 1 && cn == -1) {
			snprintf(fn, sizeof(fn), "digits/1N");
			num = 0;
		} else if (num < 20) {
			snprintf(fn, sizeof(fn), "digits/%d", num);
			num = 0;
		} else if (num < 100) {
			int ones = num % 10;
			if (ones) {
				snprintf(fn, sizeof(fn), "digits/%d-and", ones);
				num -= ones;
			} else {
				snprintf(fn, sizeof(fn), "digits/%d", num);
				num = 0;
			}
		} else {
			if (num < ONE_THOUSAND) {
				int hundreds = num / 100;
				if (hundreds == 1)
					snprintf(fn, sizeof(fn), "digits/1N");
				else
					snprintf(fn, sizeof(fn), "digits/%d", (num / 100));

				playh++;
				num -= 100 * hundreds;
				if (num)
					playa++;

			} else if (num < ONE_MILLION) {
				res = f(a, num / ONE_THOUSAND, "n");
				if (res)
					return res;
				num = num % ONE_THOUSAND;
				snprintf(fn, sizeof(fn), "digits/thousand");
			} else if (num < ONE_BILLION) {
				int millions = num / ONE_MILLION;
				res = f(a, millions, "c");
				if (res)
					return res;
				if (millions == 1)
					snprintf(fn, sizeof(fn), "digits/million");
				else
					snprintf(fn, sizeof(fn), "digits/millions");
				num = num % ONE_MILLION;
			} else {
				ast_log(LOG_DEBUG, "Number '%d' is too big for me\n", num);
				res = -1;
			}
			if (num && num < 100)
					playa++;
		}
		if (!res)
			res = s_streamwait3(a, fn);
	}
	return res;
}

/*! \brief  say_enumeration_full_da: Danish syntax */
static int say_enumeration_full_da(say_args_t *a, int num, const char *options)
{
	/* options can be: '' or 'm' male gender; 'f' female gender; 'n' neuter gender; 'p' plural */
	int res = 0, t = 0;
	char fn[256] = "", fna[256] = "";
	char *gender;

	if (options && !strncasecmp(options, "f",1)) {
		gender = "F";
	} else if (options && !strncasecmp(options, "n",1)) {
		gender = "N";
	} else {
		gender = "";
	}

	if (!num) 
		return say_digits_full(a, 0);

	while(!res && num) {
		if (num < 0) {
			snprintf(fn, sizeof(fn), "digits/minus"); /* kind of senseless for enumerations, but our best effort for error checking */
			num = negate(num);
		} else if (num < 100 && t) {
			snprintf(fn, sizeof(fn), "digits/and");
			t = 0;
		} else if (num < 20) {
			snprintf(fn, sizeof(fn), "digits/h-%d%s", num, gender);
			num = 0;
		} else if (num < 100) {
			int ones = num % 10;
			if (ones) {
				snprintf(fn, sizeof(fn), "digits/%d-and", ones);
				num -= ones;
			} else {
				snprintf(fn, sizeof(fn), "digits/h-%d%s", num, gender);
				num = 0;
			}
		} else if (num == ONE_HUNDRED && t == 0) {
			snprintf(fn, sizeof(fn), "digits/h-hundred%s", gender);
			num = 0;
		} else if (num < ONE_THOUSAND) {
			int hundreds = num / ONE_HUNDRED;
			num = num % ONE_HUNDRED;
			if (hundreds == 1) {
				snprintf(fn, sizeof(fn), "digits/1N");
			} else {
				snprintf(fn, sizeof(fn), "digits/%d", hundreds);
			}
			if (num) {					
				snprintf(fna, sizeof(fna), "digits/hundred");
			} else {
				snprintf(fna, sizeof(fna), "digits/h-hundred%s", gender);
			}
			t = 1;
		} else 	if (num < ONE_MILLION) {
			int thousands = num / ONE_THOUSAND;
			num = num % ONE_THOUSAND;
			if (thousands == 1) {
				if (num) {					
					snprintf(fn, sizeof(fn), "digits/1N");
					snprintf(fna, sizeof(fna), "digits/thousand");
				} else {
					if (t) {
						snprintf(fn, sizeof(fn), "digits/1N");
						snprintf(fna, sizeof(fna), "digits/h-thousand%s", gender);
					} else {
						snprintf(fn, sizeof(fn), "digits/h-thousand%s", gender);
					}
				}
			} else {
				res = say_number_full_da(a, thousands, options);
				if (res) {
					return res;
				}
				if (num) {					
					snprintf(fn, sizeof(fn), "digits/thousand");
				} else {
					snprintf(fn, sizeof(fn), "digits/h-thousand%s", gender);
				}
			}
			t = 1;
		} else if (num < ONE_BILLION) {
			int millions = num / ONE_MILLION;
			num = num % ONE_MILLION;
			if (millions == 1) {
				if (num) {					
					snprintf(fn, sizeof(fn), "digits/1F");
					snprintf(fna, sizeof(fna), "digits/million");
				} else {
					snprintf(fn, sizeof(fn), "digits/1N");
					snprintf(fna, sizeof(fna), "digits/h-million%s", gender);
				}
			} else {
				res = say_number_full_da(a, millions, options);
				if (res) {
					return res;
				}
				if (num) {					
					snprintf(fn, sizeof(fn), "digits/millions");
				} else {
					snprintf(fn, sizeof(fn), "digits/h-million%s", gender);
				}
			}
			t = 1;
		} else if (num < INT_MAX) {
			int billions = num / ONE_BILLION;
			num = num % ONE_BILLION;
			if (billions == 1) {
				if (num) {
					snprintf(fn, sizeof(fn), "digits/1F");
					snprintf(fna, sizeof(fna), "digits/milliard");
				} else {
					snprintf(fn, sizeof(fn), "digits/1N");
					snprintf(fna, sizeof(fna), "digits/h-milliard%s", gender);
				}
			} else {
				res = say_number_full_da(a, billions, options);
				if (res)
					return res;
				if (num) {					
					snprintf(fn, sizeof(fna), "digits/milliards");
				} else {
					snprintf(fn, sizeof(fna), "digits/h-milliard%s", gender);
				}
			}
			t = 1;
		} else if (num == INT_MAX) {
			snprintf(fn, sizeof(fn), "digits/h-last%s", gender);
			num = 0;
		} else {
			ast_log(LOG_DEBUG, "Number '%d' is too big for me\n", num);
			res = -1;
		}

		if (!res) {
			res = s_streamwait3(a, fn);
			if (!res) {
				if (!ast_strlen_zero(fna))
					res = s_streamwait3(a, fna);
				strcpy(fna, "");
			}
		}
	}
	return res;
}

/* Danish syntax */
static int say_date_with_format_da(say_args_t *a, struct tm *tm1,
	const char *format, time_t delta)
{
	struct tm tm = *tm1;
	int res=0, offset;
	char c, nextmsg[256];
	say_date_with_format_f *f = say_date_with_format_da; /* shorthand for recursive call */

	for (offset=0 ; res == 0 && (c = format[offset]) != '\0' ; offset++) {
		int year;

		ast_log(LOG_DEBUG, "Parsing %c (offset %d) in %s\n", format[offset], offset, format);
		switch (c) {
		case '\'': /* Literal name of a sound file */
			res = play_literal(a->chan, a->ints, a->language, format, &offset);
			break;
		case 'A':
		case 'a':
			/* Sunday - Saturday */
			snprintf(nextmsg,sizeof(nextmsg), "digits/day-%d", tm.tm_wday);
			res = s_waitfile(a, nextmsg);
			break;
		case 'B':
		case 'b':
		case 'h':
			/* January - December */
			snprintf(nextmsg,sizeof(nextmsg), "digits/mon-%d", tm.tm_mon);
			res = s_waitfile(a, nextmsg);
			break;
		case 'm':
			/* Month enumerated */
			res = say_enumeration_full_da(a, tm.tm_mon + 1, "m");	
			break;
		case 'd':
		case 'e':
			/* First - Thirtyfirst */
			res = say_enumeration_full_da(a, tm.tm_mday, "m");	
			break;
		case 'Y': /* Year */
			year = tm.tm_year + 1900;
			if (year > 1999) {	/* year 2000 and later */
				res = say_number_full_da(a, year, NULL);	
			} else if (year < 1100) {
				/* I'm not going to handle 1100 and prior */
				/* We'll just be silent on the year, instead of bombing out. */
			} else {
				/* year 1100 to 1999. will anybody need this?!? */
				/* say 1967 as 'nineteen hundred seven and sixty' */
				int y1 = year % ONE_HUNDRED;
				res = say_number_full_da(a, year - y1, NULL);

				if (!res && y1 != 0)
					res = say_number_full_da(a, y1, NULL);
			}
			break;
		case 'I':
		case 'l':
			/* 12-Hour */
			res = s_waitfile(a, "digits/oclock");
			if (!res)
				res = say_number_full_da(a, hr_to_12(tm.tm_hour), NULL);
			break;
		case 'H':
		case 'k':
			/* 24-Hour */
			res = s_waitfile(a, "digits/oclock");
			if (!res)
				res = say_number_full_da(a, tm.tm_hour, NULL);
			break;
		case 'M':
			/* Minute */
			if (tm.tm_min > 0 || format[offset+ 1 ] == 'S' ) {
				/* zero 'digits/0' only if seconds follow (kind of a hack) */
				res = say_number_full_da(a, tm.tm_min, "f");	
			}
			if ( !res && format[offset + 1] == 'S' ) {
				/* minutes only if seconds follow (kind of a hack) */
				res = s_waitfile(a,
					tm.tm_min == 1 ? "digits/minute" : "digits/minutes");
			}
			break;
		case 'P':
		case 'p':
			/* AM/PM */
			res = s_waitfile(a, hr_to_am_pm(tm.tm_hour));
			break;
		case 'Q': /* Short for "Today", "Yesterday", or AdBY */
		case 'q': /* Short for "" (today), "Yesterday", A (weekday), or AdBY */
			if (delta >= 0) { /* Today */
				if (c == 'Q')
					res = s_waitfile(a, "digits/today");
			} else if (delta >= - ONE_DAY) { /* Yesterday */
				res = s_waitfile(a, "digits/yesterday");
			} else if (c == 'q' && delta >=  - SIX_DAYS) {
				res = f(a, &tm, "A", delta);
			} else {
				res = f(a, &tm, "AdBY", delta);
			}
			break;
		case 'R':
			res = f(a, &tm, "HM", delta);
			break;
		case 'S':
			/* Seconds */
			res = s_waitfile(a, "digits/and");
			if (!res) {
				res = say_number_full_da(a, tm.tm_sec, "f");	
				if (!res)
					res = s_waitfile(a, "digits/seconds");
			}
			break;
		case 'T':
			res = f(a, &tm, "HMS", delta);
			break;
		case ' ':
		case '\t':
			/* Just ignore spaces and tabs */
			break;
		default:
			/* Unknown character */
			ast_log(LOG_WARNING, "Unknown character in datetime format %s: %c at pos %d\n", format, format[offset], offset);
		}
	}
	return res;
}

static struct _lang_sw map_da = {
	.language = "da",
	.say_number_full = say_number_full_da,
	.say_enumeration_full = say_enumeration_full_da,
	.say_date_with_format = say_date_with_format_da,
	.date_fmt = "AdBY",
	.time_fmt = NULL, /* XXX */
	.datetime_fmt =  NULL, /* XXX */
};

